@@ -1,3 +1,5 @@
+from __future__ import unicode_literals
+
 from gettext import gettext as _
 from . import util
 
@@ -12,19 +14,23 @@ import struct
 import termios
 import signal
 
-# This is required for ncurses to display non-ASCII characters in default user
-# locale encoding correctly.  --immerrr
-import locale
-locale.setlocale(locale.LC_ALL, u'')
-
 from .crpatch import patch, uiheader, uihunk, uihunkline
 
-# os.name is one of: 'posix', 'nt', 'dos', 'os2', 'mac', or 'ce'
-if os.name == 'posix':
+try:
     import curses
-else:
+    import fcntl
+    import termios
+    curses.error
+    fcntl.ioctl
+    termios.TIOCGWINSZ
+except ImportError:
     # I have no idea if wcurses works with crecord...
-    import wcurses as curses
+    try:
+        import wcurses as curses
+        curses.error
+    except ImportError:
+        # wcurses is not shipped on Windows by default
+        pass
 
 try:
     curses
@@ -56,7 +62,37 @@ def chunkselector(opts, headerlist, ui):
 
     """
     chunkselector = CursesChunkSelector(headerlist, ui)
+    # This is required for ncurses to display non-ASCII characters in default user
+    # locale encoding correctly.  --immerrr
+    import locale
+    locale.setlocale(locale.LC_ALL, '')
+
+    class dummystdscr(object):
+        def clear(self):
+            pass
+        def refresh(self):
+            pass
+
+    chunkselector.stdscr = dummystdscr()
+
+    f = signal.getsignal(signal.SIGTSTP)
     curses.wrapper(chunkselector.main, opts)
+    if chunkselector.initerr is not None:
+        raise util.Abort(chunkselector.initerr)
+    # ncurses does not restore signal handler for SIGTSTP
+    signal.signal(signal.SIGTSTP, f)
+
+_headermessages = { # {operation: text}
+    'crecord': _('Select hunks to commit'),
+    'cstage': _('Select hunks to stage'),
+    'cunstage': _('Select hunks to keep'),
+}
+
+_confirmmessages = {
+    'crecord': _('Are you sure you want to commit the selected changes [Yn]?'),
+    'cstage': _('Are you sure you want to stage the selected changes [Yn]?'),
+    'cunstage': _('Are you sure you want to unstage the unselected changes [Yn]?'),
+}
 
 class CursesChunkSelector(object):
     def __init__(self, headerlist, ui):
@@ -65,6 +101,7 @@ class CursesChunkSelector(object):
 
         self.ui = ui
 
+        self.errorstr = None
         # list of all chunks
         self.chunklist = []
         for h in headerlist:
@@ -77,6 +114,7 @@ class CursesChunkSelector(object):
         # maps custom nicknames of color-pairs to curses color-pair values
         self.colorpairnames = {}
 
+        self.usecolor = True
         # the currently selected header, hunk, or hunk-line
         self.currentselecteditem = self.headerlist[0]
 
@@ -96,7 +134,7 @@ class CursesChunkSelector(object):
         # keeps track of the number of lines in the pad
         self.numpadlines = None
 
-        self.numstatuslines = 2
+        self.numstatuslines = 1
 
         # keep a running count of the number of lines printed to the pad
         # (used for determining when the selected item begins/ends)
@@ -108,6 +146,45 @@ class CursesChunkSelector(object):
         # if the last 'toggle all' command caused all changes to be applied
         self.waslasttoggleallapplied = True
 
+    def handlefirstlineevent(self):
+        """
+        Handle 'g' to navigate to the top most file in the ncurses window.
+        """
+        self.currentselecteditem = self.headerlist[0]
+        currentitem = self.currentselecteditem
+        # select the parent item recursively until we're at a header
+        while True:
+            nextitem = currentitem.parentitem()
+            if nextitem is None:
+                break
+            else:
+                currentitem = nextitem
+
+        self.currentselecteditem = currentitem
+
+    def handlelastlineevent(self):
+        """
+        Handle 'G' to navigate to the bottom most file/hunk/line depending
+        on the whether the fold is active or not.
+
+        If the bottom most file is folded, it navigates to that file and stops there.
+        If the bottom most file is unfolded, it navigates to the bottom most hunk in
+        that file and stops there. If the bottom most hunk is unfolded, it navigates to
+        the bottom most line in that hunk.
+        """
+        currentitem = self.currentselecteditem
+        nextitem = currentitem.nextitem()
+        # select the child item recursively until we're at a footer
+        while nextitem is not None:
+            nextitem = currentitem.nextitem()
+            if nextitem is None:
+                break
+            else:
+                currentitem = nextitem
+
+        self.currentselecteditem = currentitem
+        self.recenterdisplayedarea()
+
     def uparrowevent(self):
         """
         Try to select the previous item to the current item that has the
@@ -126,6 +203,7 @@ class CursesChunkSelector(object):
             nextitem = currentitem
 
         self.currentselecteditem = nextitem
+        self.recenterdisplayedarea()
 
     def uparrowshiftevent(self):
         """
@@ -144,6 +222,7 @@ class CursesChunkSelector(object):
             nextitem = currentitem
 
         self.currentselecteditem = nextitem
+        self.recenterdisplayedarea()
 
     def downarrowevent(self):
         """
@@ -162,6 +241,7 @@ class CursesChunkSelector(object):
             nextitem = currentitem
 
         self.currentselecteditem = nextitem
+        self.recenterdisplayedarea()
 
     def downarrowshiftevent(self):
         """
@@ -183,6 +263,7 @@ class CursesChunkSelector(object):
             nextitem = currentitem
 
         self.currentselecteditem = nextitem
+        self.recenterdisplayedarea()
 
     def rightarrowevent(self):
         """
@@ -201,6 +282,7 @@ class CursesChunkSelector(object):
             nextitem = currentitem
 
         self.currentselecteditem = nextitem
+        self.recenterdisplayedarea()
 
     def leftarrowevent(self):
         """
@@ -227,6 +309,7 @@ class CursesChunkSelector(object):
                 self.togglefolded(item=nextitem)
 
         self.currentselecteditem = nextitem
+        self.recenterdisplayedarea()
 
     def leftarrowshiftevent(self):
         """
@@ -250,6 +333,7 @@ class CursesChunkSelector(object):
                 currentitem = nextitem
 
         self.currentselecteditem = currentitem
+        self.recenterdisplayedarea()
 
     def updatescroll(self):
         "Scroll the screen to fully show the currently-selected"
@@ -411,14 +495,9 @@ class CursesChunkSelector(object):
         width = self.xscreensize
         # turn tabs into spaces
         instr = instr.expandtabs(4)
-        try:
-            strlen = len(unicode(encoding.fromlocal(instr), code))
-        except Exception:
-            # if text is not utf8, then assume an 8-bit single-byte encoding.
-            strlen = len(instr)
-
-        numspaces = (width - ((strlen + xstart) % width) - 1)
-        return instr + " " * numspaces + "\n"
+        strwidth = encoding.ucolwidth(instr)
+        numspaces = width - ((strwidth + xstart) % width)
+        return instr + " " * numspaces
 
     def printstring(self, window, text, fgcolor=None, bgcolor=None, pair=None,
         pairname=None, attrlist=None, towin=True, align=True, showwhtspc=False):
@@ -504,10 +583,49 @@ class CursesChunkSelector(object):
 
         # is reset to 0 at the beginning of printitem()
 
-        linesprinted = (xstart + len(t)) / self.xscreensize
+        linesprinted = (xstart + len(t)) // self.xscreensize
         self.linesprintedtopadsofar += linesprinted
         return t
 
+    def _getstatuslinesegments(self):
+        """-> [str]. return segments"""
+        selected = self.currentselecteditem.applied
+        segments = [
+            _headermessages[self.opts['operation']],
+            '-',
+            _('[x]=selected **=collapsed'),
+            _('c: confirm'),
+            _('q: abort'),
+            _('arrow keys: move/expand/collapse'),
+            _('space: deselect') if selected else _('space: select'),
+            _('?: help'),
+        ]
+        return segments
+
+    def _getstatuslines(self):
+        """() -> [str]. return short help used in the top status window"""
+        if self.errorstr is not None:
+            lines = [self.errorstr, _('Press any key to continue')]
+        else:
+            # wrap segments to lines
+            segments = self._getstatuslinesegments()
+            width = self.xscreensize
+            lines = []
+            lastwidth = width
+            for s in segments:
+                w = encoding.ucolwidth(s)
+                sep = ' ' * (1 + (s and s[0] not in '-['))
+                if lastwidth + w + len(sep) >= width:
+                    lines.append(s)
+                    lastwidth = w
+                else:
+                    lines[-1] += sep + s
+                    lastwidth += w + len(sep)
+        if len(lines) != self.numstatuslines:
+            self.numstatuslines = len(lines)
+            self.statuswin.resize(self.numstatuslines, self.xscreensize)
+        return [util.ellipsis(l, self.xscreensize - 1) for l in lines]
+
     def updatescreen(self):
         self.statuswin.erase()
         self.chunkpad.erase()
@@ -516,16 +634,13 @@ class CursesChunkSelector(object):
 
         # print out the status lines at the top
         try:
-            printstring(self.statuswin,
-                        "SELECT CHUNKS: (j/k/up/dn/pgup/pgdn) move cursor; "
-                        "(space/A) toggle hunk/all; (f)old/unfold",
-                        pairname="legend")
-            printstring(self.statuswin,
-                        " (c)ommit/(s)tage applied; (q)uit; (?) help;"
-                        "toggle (a)mend mode | [x]=hunk applied **=folded",
-                        pairname="legend")
+            for line in self._getstatuslines():
+                printstring(self.statuswin, line, pairname="legend")
+            self.statuswin.refresh()
         except curses.error:
             pass
+        if self.errorstr is not None:
+            return
 
         # print out the patch in the remaining part of the window
         try:
@@ -533,14 +648,11 @@ class CursesChunkSelector(object):
             self.updatescroll()
             self.chunkpad.refresh(self.firstlineofpadtoprint, 0,
                                   self.numstatuslines, 0,
-                                  self.yscreensize + 1 - self.numstatuslines,
+                                  self.yscreensize - self.numstatuslines,
                                   self.xscreensize)
         except curses.error:
             pass
 
-        # refresh([pminrow, pmincol, sminrow, smincol, smaxrow, smaxcol])
-        self.statuswin.refresh()
-
     def getstatusprefixstring(self, item):
         """
         Create a string to prefix a line with which indicates whether 'item'
@@ -645,7 +757,7 @@ class CursesChunkSelector(object):
             # skip remainder of output
             return outstr
 
-        # print out lines of the chunk preceeding changed-lines
+        # print out lines of the chunk preceding changed-lines
         for line in hunk.before:
             linestr = " "*(self.hunklineindentnumchars + len(checkbox)) + line
             outstr += self.printstring(self.chunkpad, linestr, towin=towin)
@@ -789,7 +901,7 @@ class CursesChunkSelector(object):
         # temporarily disable printing to windows by printstring
         patchdisplaystring = self.printitem(item, ignorefolding,
                                             recursechildren, towin=False)
-        numlines = len(patchdisplaystring) / self.xscreensize
+        numlines = len(patchdisplaystring) // self.xscreensize
         return numlines
 
     def sigwinchhandler(self, n, frame):
@@ -800,6 +912,7 @@ class CursesChunkSelector(object):
             self.statuswin.resize(self.numstatuslines, self.xscreensize)
             self.numpadlines = self.getnumlinesdisplayed(ignorefolding=True) + 1
             self.chunkpad = curses.newpad(self.numpadlines, self.xscreensize)
+
         except curses.error:
             pass
 
@@ -832,11 +945,19 @@ class CursesChunkSelector(object):
                 colorpair = self.colorpairs[(fgcolor, bgcolor)]
             else:
                 pairindex = len(self.colorpairs) + 1
-                curses.init_pair(pairindex, fgcolor, bgcolor)
-                colorpair = self.colorpairs[(fgcolor, bgcolor)] = (
-                    curses.color_pair(pairindex))
-                if name is not None:
-                    self.colorpairnames[name] = curses.color_pair(pairindex)
+                if self.usecolor:
+                    curses.init_pair(pairindex, fgcolor, bgcolor)
+                    colorpair = self.colorpairs[(fgcolor, bgcolor)] = (
+                        curses.color_pair(pairindex))
+                    if name is not None:
+                        self.colorpairnames[name] = curses.color_pair(pairindex)
+                else:
+                    cval = 0
+                    if name is not None:
+                        if name == 'selected':
+                            cval = curses.A_REVERSE
+                        self.colorpairnames[name] = cval
+                    colorpair = self.colorpairs[(fgcolor, bgcolor)] = cval
 
         # add attributes if possible
         if attrlist is None:
@@ -873,6 +994,8 @@ The following are valid keystrokes:
         PgUp/PgDn [K/J] : go to previous/next item of same type
  Right/Left-arrow [l/h] : go to child item / parent item
  Shift-Left-arrow   [H] : go to parent header / fold selected header
+                      g : go to the top
+                      G : go to the bottom
                       f : fold / unfold item, hiding/revealing its children
                       F : fold / unfold parent item and all of its ancestors
                  ctrl-l : scroll the selected line to the top of the screen
@@ -888,8 +1011,9 @@ The following are valid keystrokes:
         helplines = helplines + [" "]*(
             self.yscreensize - self.numstatuslines - len(helplines) - 1)
         try:
-            for line in helplines:
-                self.printstring(helpwin, line, pairname="legend")
+            self.printstring(helpwin, helplines[0], pairname="legend")
+            for line in helplines[1:]:
+                self.printstring(helpwin, line, pairname="normal")
         except curses.error:
             pass
         helpwin.refresh()
@@ -915,25 +1039,26 @@ The following are valid keystrokes:
     def confirmationwindow(self, windowtext):
         "Display an informational window, then wait for and return a keypress."
 
-        confirmwin = curses.newwin(self.yscreensize, 0, 0, 0)
+        lines = windowtext.split("\n")
+        confirmwin = curses.newwin(len(lines), 0, 0, 0)
         try:
-            lines = windowtext.split("\n")
             for line in lines:
                 self.printstring(confirmwin, line, pairname="selected")
         except curses.error:
             pass
-        self.stdscr.refresh()
-        confirmwin.refresh()
         try:
-            response = chr(self.stdscr.getch())
+            response = chr(confirmwin.getch())
         except ValueError:
             response = None
 
         return response
 
-    def confirmcommit(self, review=False, stage=False):
+    def confirmcommit(self, review=False):
         """Ask for 'Y' to be pressed to confirm selected. Return True if
         confirmed."""
+        if not self.opts['confirm']:
+            return True
+
         if review:
             confirmtext = (
 """If you answer yes to the following, the your currently chosen patch chunks
@@ -944,18 +1069,15 @@ close the editor without saving to accept the current patch as-is.
 NOTE: don't add/remove lines unless you also modify the range information.
       Failing to follow this rule will result in the commit aborting.
 
-Are you sure you want to review/edit and confirm the selected changes [yN]?
+Are you sure you want to review/edit and confirm the selected changes [Yn]?
 """)
-        elif stage:
-            confirmtext = (
-                "Are you sure you want to stage the selected changes [yN]? ")
         else:
-            confirmtext = (
-                "Are you sure you want to commit the selected changes [yN]? ")
+            confirmtext = _confirmmessages[self.opts['operation']]
 
         response = self.confirmationwindow(confirmtext)
-        if response is None:
-            response = "n"
+
+        if response is None or len(response) == 0 or response == "\n":
+            response = "y"
         if response.lower().startswith("y"):
             return True
         else:
@@ -965,13 +1087,13 @@ Are you sure you want to review/edit and confirm the selected changes [yN]?
         """
         once we scrolled with pg up pg down we can be pointing outside of the
         display zone. we print the patch with towin=False to compute the
-        location of the selected item eventhough it is outside of the displayed
+        location of the selected item even though it is outside of the displayed
         zone and then update the scroll.
         """
         self.printitem(towin=False)
         self.updatescroll()
 
-    def toggleamend(self, opts):
+    def toggleamend(self):
         """Toggle the amend flag.
 
         When the amend flag is set, a commit will modify the most recently
@@ -979,15 +1101,15 @@ Are you sure you want to review/edit and confirm the selected changes [yN]?
         new changeset will be created (the normal commit behavior).
 
         """
-        if opts.get('amend') is False:
-            opts['amend'] = True
-            msg = ("Amend option is turned on -- commiting the currently "
+        if self.opts.get('amend') is False:
+            self.opts['amend'] = True
+            msg = ("Amend option is turned on -- committing the currently "
                    "selected changes will not create a new changeset, but "
                    "instead update the most recently committed changeset.\n\n"
                    "Press any key to continue.")
-        elif opts.get('amend') is True:
-            opts['amend'] = False
-            msg = ("Amend option is turned off -- commiting the currently "
+        elif self.opts.get('amend') is True:
+            self.opts['amend'] = False
+            msg = ("Amend option is turned off -- committing the currently "
                    "selected changes will create a new changeset.\n\n"
                    "Press any key to continue.")
 
@@ -1002,7 +1124,7 @@ Are you sure you want to review/edit and confirm the selected changes [yN]?
                 return False
         return True
 
-    def handlekeypressed(self, keypressed, opts):
+    def handlekeypressed(self, keypressed):
         """
         Perform actions based on pressed keys.
 
@@ -1010,7 +1132,7 @@ Are you sure you want to review/edit and confirm the selected changes [yN]?
         """
         if keypressed in ["k", "KEY_UP"]:
             self.uparrowevent()
-        if keypressed in ["K", "KEY_PPAGE"]:
+        elif keypressed in ["K", "KEY_PPAGE"]:
             self.uparrowshiftevent()
         elif keypressed in ["j", "KEY_DOWN"]:
             self.downarrowevent()
@@ -1025,20 +1147,20 @@ Are you sure you want to review/edit and confirm the selected changes [yN]?
         elif keypressed in ["q"]:
             raise util.Abort(_('user quit'))
         elif keypressed in ['a']:
-            self.toggleamend(opts)
+            self.toggleamend()
         elif keypressed in ["c"]:
             if self.confirmcommit():
-                opts['commit'] = True
+                self.opts['commit'] = True
                 return True
         elif keypressed in ["s"]:
-            opts['commit'] = False
-            if self.confirmcommit(stage=True):
-                opts['commit'] = False
+            self.opts['commit'] = False
+            if self.confirmcommit():
+                self.opts['commit'] = False
                 return True
         elif keypressed in ["r"]:
             if self.confirmcommit(review=True):
-                opts['commit'] = True
-                opts['crecord_reviewpatch'] = True
+                self.opts['commit'] = True
+                self.opts['crecord_reviewpatch'] = True
                 return True
         elif keypressed in [' ']:
             self.toggleapply()
@@ -1047,14 +1169,21 @@ Are you sure you want to review/edit and confirm the selected changes [yN]?
         elif keypressed in ["f"]:
             self.togglefolded()
         elif keypressed in ["F"]:
-            self.togglefolded(foldParent=True)
+            self.togglefolded(foldparent=True)
         elif keypressed in ["?"]:
             self.helpwindow()
             self.stdscr.clear()
             self.stdscr.refresh()
-        elif curses.unctrl(keypressed) in ["^L"]:
-            # scroll the current line to the top of the screen
+        elif len(keypressed) == 1 and curses.unctrl(keypressed) in [b"^L"]:
+            # scroll the current line to the top of the screen, and redraw
+            # everything
             self.scrolllines(self.selecteditemstartline)
+            self.stdscr.clear()
+            self.stdscr.refresh()
+        elif keypressed in ["g", "KEY_HOME"]:
+            self.handlefirstlineevent()
+        elif keypressed in ["G", "KEY_END"]:
+            self.handlelastlineevent()
         return False
 
     def main(self, stdscr, opts):
@@ -1062,13 +1191,42 @@ Are you sure you want to review/edit and confirm the selected changes [yN]?
         Method to be wrapped by curses.wrapper() for selecting chunks.
 
         """
-        origsigwinchhandler = signal.signal(signal.SIGWINCH,
-                                            self.sigwinchhandler)
+        self.opts = opts
+
+        origsigwinch = sentinel = object()
+        if util.safehasattr(signal, 'SIGWINCH'):
+             origsigwinch = signal.signal(signal.SIGWINCH,
+                                          self.sigwinchhandler)
+        try:
+            return self._main(stdscr)
+        finally:
+            if origsigwinch is not sentinel:
+                signal.signal(signal.SIGWINCH, origsigwinch)
+
+    def _main(self, stdscr):
         self.stdscr = stdscr
+        # error during initialization, cannot be printed in the curses
+        # interface, it should be printed by the calling code
+        self.initerr = None
         self.yscreensize, self.xscreensize = self.stdscr.getmaxyx()
 
         curses.start_color()
-        curses.use_default_colors()
+        try:
+            curses.use_default_colors()
+        except curses.error:
+            self.usecolor = False
+
+        # In some situations we may have some cruft left on the "alternate
+        # screen" from another program (or previous iterations of ourself), and
+        # we won't clear it if the scroll region is small enough to comfortably
+        # fit on the terminal.
+        self.stdscr.clear()
+
+        # don't display the cursor
+        try:
+            curses.curs_set(0)
+        except curses.error:
+            pass
 
         # available colors: black, blue, cyan, green, magenta, white, yellow
         # init_pair(color_id, foreground_color, background_color)
@@ -1090,25 +1248,31 @@ Are you sure you want to review/edit and confirm the selected changes [yN]?
 
         # add 1 so to account for last line text reaching end of line
         self.numpadlines = self.getnumlinesdisplayed(ignorefolding=True) + 1
-        self.chunkpad = curses.newpad(self.numpadlines, self.xscreensize)
 
-        # initialize selecteitemendline (initial start-line is 0)
+        try:
+            self.chunkpad = curses.newpad(self.numpadlines, self.xscreensize)
+        except curses.error:
+            self.initerr = _('this diff is too large to be displayed')
+            return
+        # initialize selecteditemendline (initial start-line is 0)
         self.selecteditemendline = self.getnumlinesdisplayed(
             self.currentselecteditem, recursechildren=False)
 
         # option which enables/disables patch-review (in editor) step
-        opts['crecord_reviewpatch'] = False
+        self.opts['crecord_reviewpatch'] = False
 
-        if opts['author'] is not None:
+        if self.opts['author'] is not None:
             # make it accessible by self.ui.username()
-            self.ui.setusername(opts['author'])
+            self.ui.setusername(self.opts['author'])
 
         while True:
             self.updatescreen()
             try:
                 keypressed = self.statuswin.getkey()
+                if self.errorstr is not None:
+                    self.errorstr = None
+                    continue
             except curses.error:
                 keypressed = "FOOBAR"
-            if self.handlekeypressed(keypressed, opts):
+            if self.handlekeypressed(keypressed):
                 break
-        signal.signal(signal.SIGWINCH, origsigwinchhandler)
