@@ -1,7 +1,9 @@
+from __future__ import unicode_literals
+
 # stuff related specifically to patch manipulation / parsing
 from gettext import gettext as _
 
-import cStringIO
+import io
 import re
 
 class PatchError(Exception):
@@ -24,14 +26,10 @@ class linereader(object):
             l = self.buf[0]
             del self.buf[0]
             return l
-        return self.fp.readline()
+        return self.fp.readline().decode('UTF-8')
 
     def __iter__(self):
-        while True:
-            l = self.readline()
-            if not l:
-                break
-            yield l
+        return iter(self.readline, '')
 
 def scanpatch(fp):
     """like patch.iterhunks, but yield different events
@@ -46,10 +44,7 @@ def scanpatch(fp):
     def scanwhile(first, p):
         """scan lr while predicate holds"""
         lines = [first]
-        while True:
-            line = lr.readline()
-            if not line:
-                break
+        for line in iter(lr.readline, ''):
             if p(line):
                 lines.append(line)
             else:
@@ -57,10 +52,7 @@ def scanpatch(fp):
                 break
         return lines
 
-    while True:
-        line = lr.readline()
-        if not line:
-            break
+    for line in iter(lr.readline, ''):
         if line.startswith('diff --git a/'):
             def notheader(line):
                 s = line.split(None, 1)
@@ -73,7 +65,7 @@ def scanpatch(fp):
             else:
                 lr.push(fromfile)
             yield 'file', header
-        elif line[0] == ' ':
+        elif line.startswith(' '):
             yield 'context', scanwhile(line, lambda l: l[0] in ' \\')
         elif line[0] in '-+':
             yield 'hunk', scanwhile(line, lambda l: l[0] in '-+\\')
@@ -238,10 +230,7 @@ class uiheader(patchnode):
         Otherwise return False.
 
         """
-        for h in self.header:
-            if h.startswith('GIT binary patch'):
-                return True
-        return False
+        return any(h.startswith('GIT binary patch') for h in self.header)
 
     def pretty(self, fp):
         for h in self.header:
@@ -256,12 +245,12 @@ class uiheader(patchnode):
             if h.startswith('---'):
                 fp.write(_('%d hunks, %d lines changed\n') %
                          (len(self.hunks),
-                          sum([h.added + h.removed for h in self.hunks])))
+                          sum([max(h.added, h.removed) for h in self.hunks])))
                 break
             fp.write(h)
 
     def prettystr(self):
-        x = cStringIO.StringIO()
+        x = io.StringIO()
         self.pretty(x)
         return x.getvalue()
 
@@ -275,10 +264,7 @@ class uiheader(patchnode):
         smaller than the size of the entire file.)  Otherwise return False
 
         """
-        for h in self.header:
-            if self.allhunks_re.match(h):
-                return True
-        return False
+        return any(self.allhunks_re.match(h) for h in self.header)
 
     def files(self):
         fromfile, tofile = self.diff_re.match(self.header[0]).groups()
@@ -296,9 +282,7 @@ class uiheader(patchnode):
         return '<header %s>' % (' '.join(map(repr, self.files())))
 
     def special(self):
-        for h in self.header:
-            if self.special_re.match(h):
-                return True
+        return any(self.special_re.match(h) for h in self.header)
 
     @property
     def changetype(self):
@@ -476,9 +460,9 @@ class uihunk(patchnode):
     def countchanges(self):
         """changedlines -> (n+,n-)"""
         add = len([l for l in self.changedlines if l.applied
-                   and l.prettystr()[0] == '+'])
+                   and l.prettystr().startswith('+')])
         rem = len([l for l in self.changedlines if l.applied
-                   and l.prettystr()[0] == '-'])
+                   and l.prettystr().startswith('-')])
         return add, rem
 
     def getfromtoline(self):
@@ -499,12 +483,11 @@ class uihunk(patchnode):
         # So, if either of hunks is empty, decrease its line start. --immerrr
         # But only do this if fromline > 0, to avoid having, e.g fromline=-1.
         fromline, toline = self.fromline, self.toline
-        if fromlen == 0:
-            if fromline != 0:
+        if fromline != 0:
+            if fromlen == 0:
                 fromline -= 1
-        if tolen == 0:
-            if toline != 0:
-                toline -= 1
+        if tolen == 0 and toline > 0:
+            toline -= 1
 
         fromtoline = '@@ -%d,%d +%d,%d @@%s\n' % (
             fromline, fromlen, toline, tolen,
@@ -523,18 +506,35 @@ class uihunk(patchnode):
             changedlinestr = changedline.prettystr()
             if changedline.applied:
                 hunklinelist.append(changedlinestr)
-            elif changedlinestr[0] == "-":
+            elif changedlinestr.startswith("-"):
                 hunklinelist.append(" " + changedlinestr[1:])
 
         fp.write(''.join(self.before + hunklinelist + self.after))
 
+    def reversehunks(self):
+       m = {'+': '-', '-': '+', '\\': '\\'}
+       hunk = ['%s%s' % (m[l.prettystr()[0:1]], l.prettystr()[1:]) for l in self.changedlines if l.applied]
+       return uihunk(self.header, self.fromline, self.toline, self.proc, self.before, hunk, self.after)
+
+    def unapplyhunks(self):
+        m = {'+': '-', '-': '+', '\\': '\\'}
+        hunklinelist = []
+        for changedline in self.changedlines:
+           changedlinestr = changedline.prettystr()
+           if not changedline.applied:
+               hunklinelist.append('%s%s' % (m[changedlinestr[0]], changedlinestr[1:]))
+           elif changedlinestr.startswith("+"):
+               hunklinelist.append(" " + changedlinestr[1:])
+        return uihunk(self.header, self.fromline, self.toline, self.proc, self.before, hunklinelist, self.after)
+
+
     pretty = write
 
     def filename(self):
         return self.header.filename()
 
     def prettystr(self):
-        x = cStringIO.StringIO()
+        x = io.StringIO()
         self.pretty(x)
         return x.getvalue()
 
@@ -554,11 +554,12 @@ def parsepatch(fp):
             self.header = None
             self.context = []
             self.before = []
-            self.changedlines = []
-            self.stream = []
+            self.hunk = []
+            self.headers = []
 
-        def _range(self, (fromstart, fromend, tostart, toend, proc)):
+        def addrange(self, limits):
             "Store range line info to associated instance variables."
+            fromstart, fromend, tostart, toend, proc = limits
             self.fromline = int(fromstart)
             self.toline = int(tostart)
             self.proc = proc
@@ -566,7 +567,7 @@ def parsepatch(fp):
         def add_new_hunk(self):
             """
             Create a new complete hunk object, adding it to the latest header
-            and to self.stream.
+            and to self.headers.
 
             Add all of the previously collected information about
             the hunk to the new hunk object.  This information includes
@@ -579,17 +580,17 @@ def parsepatch(fp):
 
             """
             h = uihunk(self.header, self.fromline, self.toline, self.proc,
-                       self.before, self.changedlines, self.context)
+                       self.before, self.hunk, self.context)
             self.header.hunks.append(h)
-            self.stream.append(h)
+            self.headers.append(h)
             self.fromline += len(self.before) + h.removed + len(self.context)
             self.toline += len(self.before) + h.added + len(self.context)
             self.before = []
-            self.changedlines = []
+            self.hunk = []
             self.context = []
             self.proc = ''
 
-        def _context(self, context):
+        def addcontext(self, context):
             """
             Set the value of self.context.
 
@@ -602,10 +603,10 @@ def parsepatch(fp):
             self.context = context
             # if there have been changed lines encountered that haven't yet
             # been add to a hunk.
-            if self.changedlines:
+            if self.hunk:
                 self.add_new_hunk()
 
-        def _changedlines(self, changedlines):
+        def addhunk(self, hunk):
             """
             Store the changed lines in self.changedlines.
 
@@ -614,48 +615,47 @@ def parsepatch(fp):
             clear the context-line buffer.
 
             """
-            self.changedlines = changedlines
+            self.hunk = hunk
             self.before = self.context
             self.context = []
 
-        def add_new_header(self, hdr):
+        def newfile(self, hdr):
             """
             Create a header object containing the header lines, and the
-            filename the header applies to.  Add the header to self.stream.
+            filename the header applies to.  Add the header to self.headers.
 
             """
             # if there are any lines in the unchanged-lines buffer, create a 
             # new hunk using them, and add it to the last header.
-            if self.changedlines:
+            if self.hunk:
                 self.add_new_hunk()
 
-            # create a new header and add it to self.stream
-            self.header = uiheader(hdr)
-            fileName = self.header.filename()
-
-            self.stream.append(self.header)
+            # create a new header and add it to self.header
+            h = uiheader(hdr)
+            self.headers.append(h)
+            self.header = h
 
         def finished(self):
             # if there are any lines in the unchanged-lines buffer, create a 
             # new hunk using them, and add it to the last header.
-            if self.changedlines:
+            if self.hunk:
                 self.add_new_hunk()
 
-            return self.stream
+            return self.headers
 
         transitions = {
-            'file': {'context': _context,
-                     'file': add_new_header,
-                     'hunk': _changedlines,
-                     'range': _range},
-            'context': {'file': add_new_header,
-                        'hunk': _changedlines,
-                        'range': _range},
-            'hunk': {'context': _context,
-                     'file': add_new_header,
-                     'range': _range},
-            'range': {'context': _context,
-                      'hunk': _changedlines},
+            'file': {'context': addcontext,
+                     'file': newfile,
+                     'hunk': addhunk,
+                     'range': addrange},
+            'context': {'file': newfile,
+                        'hunk': addhunk,
+                        'range': addrange},
+            'hunk': {'context': addcontext,
+                     'file': newfile,
+                     'range': addrange},
+            'range': {'context': addcontext,
+                      'hunk': addhunk},
             }
 
     p = parser()
@@ -688,7 +688,7 @@ def filterpatch(opts, chunks, chunkselector, ui):
     appliedHunkList = []
     for hdr in headers:
         if (hdr.applied and
-            (hdr.special() or len([h for h in hdr.hunks if h.applied]) > 0)):
+            (hdr.special() or hdr.binary() or len([h for h in hdr.hunks if h.applied]) > 0)):
             appliedHunkList.append(hdr)
             fixoffset = 0
             for hnk in hdr.hunks:
